@@ -76,6 +76,8 @@ from subprocess import (
     Popen,
     PIPE,
 )
+from textwrap import wrap
+
 import apt
 import apt_inst
 import apt_pkg
@@ -150,25 +152,16 @@ class UnattendedUpgradesCache(apt.Cache):
         datetime.datetime.strptime("", "")
 
         # generate versioned_kernel_pkgs_regexp for later use
-        apt_versioned_kernel_pkgs = apt_pkg.config.value_list(
-            "APT::VersionedKernelPackages")
-        if apt_versioned_kernel_pkgs:
-            self.versioned_kernel_pkgs_regexp = re.compile("(" + "|".join(
-                ["^" + p + "-[0-9]+\\.[0-9\\.]+-.*"
-                 for p in apt_versioned_kernel_pkgs]) + ")")
+        self.versioned_kernel_pkgs_regexp = versioned_kernel_pkgs_regexp()
+        self.running_kernel_pkgs_regexp = running_kernel_pkgs_regexp()
+        if self.versioned_kernel_pkgs_regexp:
             logging.debug("Using %s regexp to find kernel packages",
                           self.versioned_kernel_pkgs_regexp.pattern)
-            running_kernel_version = subprocess.check_output(
-                ["uname", "-r"], universal_newlines=True).rstrip()
-            self.running_kernel_pkgs_regexp = re.compile("(" + "|".join(
-                [("^" + p + "-" + re.escape(running_kernel_version) + "$")
-                 for p in apt_versioned_kernel_pkgs]) + ")")
-            logging.debug("Using %s regexp to find running kernel packages",
-                          self.running_kernel_pkgs_regexp.pattern)
         else:
             logging.debug("APT::VersionedKernelPackages is not set")
-            self.versioned_kernel_pkgs_regexp = None
-            self.running_kernel_pkgs_regexp = None
+        if self.running_kernel_pkgs_regexp:
+            logging.debug("Using %s regexp to find running kernel packages",
+                          self.running_kernel_pkgs_regexp.pattern)
 
     def adjust_candidate(self, pkg):
         # type: (apt.Package) -> bool
@@ -575,6 +568,41 @@ def get_distro_id():
     return DISTRO_ID
 
 
+def versioned_kernel_pkgs_regexp():
+    apt_versioned_kernel_pkgs = apt_pkg.config.value_list(
+        "APT::VersionedKernelPackages")
+    if apt_versioned_kernel_pkgs:
+        return re.compile("(" + "|".join(
+            ["^" + p + "-[1-9][0-9]*\\.[0-9]+\\.[0-9]+-[0-9]+(-.+)?$"
+             for p in apt_versioned_kernel_pkgs]) + ")")
+    else:
+        return None
+
+
+def running_kernel_pkgs_regexp():
+    apt_versioned_kernel_pkgs = apt_pkg.config.value_list(
+        "APT::VersionedKernelPackages")
+    if apt_versioned_kernel_pkgs:
+        running_kernel_version = subprocess.check_output(
+            ["uname", "-r"], universal_newlines=True).rstrip()
+        kernel_escaped = re.escape(running_kernel_version)
+        try:
+            kernel_noflavor_escaped = re.escape(
+                re.match("[1-9][0-9]*\\.[0-9]+\\.[0-9]+-[0-9]+",
+                         running_kernel_version)[0])
+            return re.compile("(" + "|".join(
+                [("^" + p + "-" + kernel_escaped + "$|^"
+                  + p + "-" + kernel_noflavor_escaped + "$")
+                 for p in apt_versioned_kernel_pkgs]) + ")")
+        except TypeError:
+            # flavor could not be cut from version
+            return re.compile("(" + "|".join(
+                [("^" + p + "-" + kernel_escaped + "$")
+                 for p in apt_versioned_kernel_pkgs]) + ")")
+    else:
+        return None
+
+
 def get_allowed_origins_legacy():
     # type: () -> List[str]
     """ legacy support for old Allowed-Origins var """
@@ -931,6 +959,8 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
                 logging.debug("pkg %s not in allowed origin" % pkg.name)
                 return False
             if not is_pkg_change_allowed(pkg, blacklist, whitelist):
+                logging.debug("pkg %s is blacklisted or not whitelisted"
+                              % pkg.name)
                 return False
             # check if the package is unsafe to upgrade unattended
             ignore_require_restart = apt_pkg.config.find_b(
@@ -943,6 +973,8 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
                 return False
     # check that the package we want to upgrade is in the change set
     if desired_pkg and desired_pkg not in changes:
+        logging.debug("pkg %s to be marked for upgrade/install is not marked "
+                      "accordingly" % pkg.name)
         return False
     return True
 
@@ -1154,16 +1186,10 @@ def host():
     return socket.getfqdn()
 
 
-# *sigh* textwrap is nice, but it breaks "linux-image" into two
-# seperate lines
-def wrap(t, width=70, subsequent_indent=""):
-    # type: (str, int, str) -> str
-    out = ""
-    for s in t.split():
-        if (len(out) - out.rfind("\n")) + len(s) > width:
-            out += "\n" + subsequent_indent
-        out += s + " "
-    return out
+def wrap_indent(t, subsequent_indent=" "):
+    # type: (str, str) -> str
+    return "\n".join(wrap(t, break_on_hyphens=False,
+                          subsequent_indent=subsequent_indent))
 
 
 def setup_apt_listchanges(conf="/etc/apt/listchanges.conf"):
@@ -1235,18 +1261,36 @@ def send_summary_mail(pkgs,                 # type: List[str]
     if not to_email:
         return
     if not os.path.exists(MAIL_BINARY) and not os.path.exists(SENDMAIL_BINARY):
-        logging.error(_("No /usr/bin/mail or /usr/sbin/sendmail,"
+        logging.error(_("No /usr/bin/mail or /usr/sbin/sendmail, "
                         "can not send mail. "
                         "You probably want to install the mailx package."))
         return
+
+    # The admin may well wish to get a mail report regardless of what was done.
+    # This is now set by Unattended-Upgrade::MailReport values of:
+    #       "always", "only-on-error" or "on-change"
+    # (you can achieve "never" by not setting Unattended-Upgrade::Mail).
+    # If this is not set, then set it using any legacy MailOnlyOnError
+    # setting (default True)
+    #
+    mail_opt = apt_pkg.config.find("Unattended-Upgrade::MailReport")
+    if (mail_opt == ""):    # None set - map from legacy value
+        if apt_pkg.config.find_b("Unattended-Upgrade::MailOnlyOnError", False):
+            mail_opt = "only-on-error"
+        else:
+            mail_opt = "on-change"
+
     # if the operation was successful and the user has requested to get
-    # mails on on errors, just exit here
-    if (res and apt_pkg.config.find_b(
-            "Unattended-Upgrade::MailOnlyOnError", False)):
+    # mails only on errors, just exit here
+    if (res and (mail_opt == "only-on-error")):
         return
+
     # if the run was successful but nothing had to be done skip sending email
-    if (res and not pkgs and not pkgs_kept_back and not pkgs_removed):
+    # unless the admin wants it anyway
+    if (((mail_opt != "always") and res and not pkgs and not pkgs_kept_back
+         and not pkgs_removed)):
         return
+
     # Check if reboot-required flag is present
     reboot_flag_str = _(
         "[reboot required]") if os.path.isfile(REBOOT_REQUIRED_FILE) else ""
@@ -1259,7 +1303,7 @@ def send_summary_mail(pkgs,                 # type: List[str]
         "{machine}: {result}").format(
             hold_flag=hold_flag_str, reboot_flag=reboot_flag_str,
             machine=host(), result="SUCCESS" if res else "FAILURE").strip()
-    body = wrap(_("Unattended upgrade result: %s") % result_str, 70, " ")
+    body = wrap_indent(_("Unattended upgrade result: %s") % result_str)
     body += "\n\n"
     if os.path.isfile(REBOOT_REQUIRED_FILE):
         body += _(
@@ -1270,21 +1314,22 @@ def send_summary_mail(pkgs,                 # type: List[str]
             body += _("Packages that were upgraded:\n")
         else:
             body += _("Packages that attempted to upgrade:\n")
-        body += " " + wrap(" ".join(pkgs), 70, " ")
+        body += " " + wrap_indent(" ".join(pkgs))
         body += "\n\n"
     if pkgs_kept_back:
         body += _("Packages with upgradable origin but kept back:\n")
         for origin, origin_pkgs in pkgs_kept_back.items():
             body += " " + origin + ":\n"
-            body += "  " + wrap(" ".join(origin_pkgs), 70, "  ") + "\n"
+            body += "  " + wrap_indent(" ".join(origin_pkgs),
+                                       subsequent_indent="  ") + "\n"
         body += "\n"
     if pkgs_removed:
         body += _("Packages that were auto-removed:\n")
-        body += " " + wrap(" ".join(pkgs_removed), 70, " ")
+        body += " " + wrap_indent(" ".join(pkgs_removed))
         body += "\n\n"
     if pkgs_kept_installed:
         body += _("Packages that were kept from being auto-removed:\n")
-        body += " " + wrap(" ".join(pkgs_kept_installed), 70, " ")
+        body += " " + wrap_indent(" ".join(pkgs_kept_installed))
         body += "\n\n"
     if dpkg_log_content:
         body += _("Package installation log:") + "\n"
@@ -1301,8 +1346,9 @@ def send_summary_mail(pkgs,                 # type: List[str]
         ret = _send_mail_using_mailx(from_email, to_email, subject, body)
     else:
         raise AssertionError(
-            "This should never be reached, if we are here we either "
-            "have sendmail or mailx. Please report this as a bug.")
+            "This should never be reached as we previously validated that we "
+            "either have sendmail or mailx. Maybe they've been removed in "
+            "this right moment?")
     logging.debug("mail returned: %s", ret)
 
 
@@ -1651,6 +1697,12 @@ def is_autoremove_valid(cache, pkgname, auto_removable, blacklist, whitelist):
                   "they include %s which package is related to the running "
                   "kernel: %s"), packagename, " ".join(sorted(pkgnames)))
             return False
+    if cache.install_count > 0:
+        logging.error(
+            "The following packages are marked for installation or upgrade "
+            "which is not allowed when performing autoremovals: %s",
+            " ".join([pkg.name for pkg in changes if not pkg.marked_delete]))
+        return False
     return True
 
 
